# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import string

import jmespath
import pytest
from jmespath.exceptions import JMESPathError

import botocore.session

KNOWN_PAGE_KEYS = {
    'input_token',
    'py_input_token',
    'output_token',
    'result_key',
    'limit_key',
    'more_results',
    'non_aggregate_keys',
}
MEMBER_NAME_CHARS = set(string.ascii_letters + string.digits)
# The goal here should be to remove all of these by updating the paginators
# to reference all the extra output keys. Nothing should ever be added to this
# list, it represents all the current released paginators that fail this test.
KNOWN_EXTRA_OUTPUT_KEYS = [
    'alexaforbusiness.SearchUsers.TotalCount',
    'alexaforbusiness.SearchProfiles.TotalCount',
    'alexaforbusiness.SearchSkillGroups.TotalCount',
    'alexaforbusiness.SearchDevices.TotalCount',
    'alexaforbusiness.SearchRooms.TotalCount',
    'apigateway.GetApiKeys.warnings',
    'apigateway.GetUsage.usagePlanId',
    'apigateway.GetUsage.startDate',
    'apigateway.GetUsage.endDate',
    'athena.GetQueryResults.ResultSet',
    'cloudfront.ListCloudFrontOriginAccessIdentities.CloudFrontOriginAccessIdentityList',
    'cloudfront.ListDistributions.DistributionList',
    'cloudfront.ListInvalidations.InvalidationList',
    'cloudfront.ListStreamingDistributions.StreamingDistributionList',
    'cloudfront.ListKeyValueStores.KeyValueStoreList',
    'codedeploy.ListDeploymentGroups.applicationName',
    'dms.DescribeTableStatistics.ReplicationTaskArn',
    'dms.DescribeReplicationTaskAssessmentResults.BucketName',
    'ec2.DescribeSpotFleetInstances.SpotFleetRequestId',
    'ec2.DescribeVpcEndpointServices.ServiceNames',
    'efs.DescribeFileSystems.Marker',
    'efs.DescribeMountTargets.Marker',
    'efs.DescribeTags.Marker',
    'elasticache.DescribeCacheParameters.CacheNodeTypeSpecificParameters',
    'elasticache.DescribeEngineDefaultParameters.EngineDefaults',
    'glacier.ListParts.PartSizeInBytes',
    'glacier.ListParts.ArchiveDescription',
    'glacier.ListParts.MultipartUploadId',
    'glacier.ListParts.VaultARN',
    'glacier.ListParts.CreationDate',
    'kinesis.DescribeStream.StreamDescription',
    'mturk.ListAssignmentsForHIT.NumResults',
    'mturk.ListQualificationTypes.NumResults',
    'mturk.ListHITs.NumResults',
    'mturk.ListWorkerBlocks.NumResults',
    'mturk.ListReviewableHITs.NumResults',
    'mturk.ListHITsForQualificationType.NumResults',
    'mturk.ListQualificationRequests.NumResults',
    'mturk.ListWorkersWithQualificationType.NumResults',
    'mturk.ListBonusPayments.NumResults',
    'neptune.DescribeEngineDefaultParameters.EngineDefaults',
    'rds.DescribeEngineDefaultClusterParameters.EngineDefaults',
    'rds.DescribeEngineDefaultParameters.EngineDefaults',
    'redshift.DescribeDefaultClusterParameters.DefaultClusterParameters',
    'resource-groups.ListGroups.GroupIdentifiers',
    'resource-groups.SearchResources.QueryErrors',
    'resource-groups.ListGroupResources.QueryErrors',
    'route53.ListHealthChecks.MaxItems',
    'route53.ListHealthChecks.Marker',
    'route53.ListHostedZones.MaxItems',
    'route53.ListHostedZones.Marker',
    'route53.ListResourceRecordSets.MaxItems',
    's3.ListMultipartUploads.Delimiter',
    's3.ListMultipartUploads.KeyMarker',
    's3.ListMultipartUploads.Prefix',
    's3.ListMultipartUploads.Bucket',
    's3.ListMultipartUploads.MaxUploads',
    's3.ListMultipartUploads.UploadIdMarker',
    's3.ListMultipartUploads.EncodingType',
    's3.ListObjectVersions.MaxKeys',
    's3.ListObjectVersions.Delimiter',
    's3.ListObjectVersions.VersionIdMarker',
    's3.ListObjectVersions.KeyMarker',
    's3.ListObjectVersions.Prefix',
    's3.ListObjectVersions.Name',
    's3.ListObjectVersions.EncodingType',
    's3.ListObjects.MaxKeys',
    's3.ListObjects.Delimiter',
    's3.ListObjects.NextMarker',
    's3.ListObjects.Prefix',
    's3.ListObjects.Marker',
    's3.ListObjects.Name',
    's3.ListObjects.EncodingType',
    's3.ListObjectsV2.StartAfter',
    's3.ListObjectsV2.MaxKeys',
    's3.ListObjectsV2.Delimiter',
    's3.ListObjectsV2.ContinuationToken',
    's3.ListObjectsV2.KeyCount',
    's3.ListObjectsV2.Prefix',
    's3.ListObjectsV2.Name',
    's3.ListObjectsV2.EncodingType',
    's3.ListParts.PartNumberMarker',
    's3.ListParts.AbortDate',
    's3.ListParts.MaxParts',
    's3.ListParts.Bucket',
    's3.ListParts.Key',
    's3.ListParts.UploadId',
    's3.ListParts.AbortRuleId',
    's3.ListParts.RequestCharged',
    'sms.GetReplicationRuns.replicationJob',
    'sms.GetServers.lastModifiedOn',
    'sms.GetServers.serverCatalogStatus',
    'storagegateway.DescribeTapeRecoveryPoints.GatewayARN',
    'storagegateway.DescribeVTLDevices.GatewayARN',
    'storagegateway.ListVolumes.GatewayARN',
    'workdocs.DescribeUsers.TotalNumberOfUsers',
    'xray.BatchGetTraces.UnprocessedTraceIds',
    'xray.GetServiceGraph.EndTime',
    'xray.GetServiceGraph.ContainsOldGroupVersions',
    'xray.GetServiceGraph.StartTime',
    'xray.GetTraceSummaries.TracesProcessedCount',
    'xray.GetTraceSummaries.ApproximateTime',
]


def _pagination_configs():
    session = botocore.session.get_session()
    loader = session.get_component('data_loader')
    services = loader.list_available_services('paginators-1')
    for service_name in services:
        service_model = session.get_service_model(service_name)
        page_config = loader.load_service_model(
            service_name, 'paginators-1', service_model.api_version
        )
        for op_name, single_config in page_config['pagination'].items():
            yield (op_name, single_config, service_model)


@pytest.mark.parametrize(
    "operation_name, page_config, service_model", _pagination_configs()
)
def test_lint_pagination_configs(operation_name, page_config, service_model):
    _validate_known_pagination_keys(page_config)
    _valiate_result_key_exists(page_config)
    _validate_referenced_operation_exists(operation_name, service_model)
    _validate_operation_has_output(operation_name, service_model)
    _validate_input_keys_match(operation_name, page_config, service_model)
    _validate_output_keys_match(operation_name, page_config, service_model)


def _validate_known_pagination_keys(page_config):
    for key in page_config:
        if key not in KNOWN_PAGE_KEYS:
            raise AssertionError(
                "Unknown key '%s' in pagination config: %s"
                % (key, page_config)
            )


def _valiate_result_key_exists(page_config):
    if 'result_key' not in page_config:
        raise AssertionError(
            "Required key 'result_key' is missing "
            "from pagination config: %s" % page_config
        )


def _validate_referenced_operation_exists(operation_name, service_model):
    if operation_name not in service_model.operation_names:
        raise AssertionError(
            "Pagination config refers to operation that "
            "does not exist: %s" % operation_name
        )


def _validate_operation_has_output(operation_name, service_model):
    op_model = service_model.operation_model(operation_name)
    output = op_model.output_shape
    if output is None or not output.members:
        raise AssertionError(
            "Pagination config refers to operation "
            "that does not have any output: %s" % operation_name
        )


def _validate_input_keys_match(operation_name, page_config, service_model):
    input_tokens = page_config['input_token']
    if not isinstance(input_tokens, list):
        input_tokens = [input_tokens]
    valid_input_names = service_model.operation_model(
        operation_name
    ).input_shape.members
    for token in input_tokens:
        if token not in valid_input_names:
            raise AssertionError(
                "input_token '%s' refers to a non existent "
                "input member for operation: %s" % (token, operation_name)
            )
    if 'limit_key' in page_config:
        limit_key = page_config['limit_key']
        if limit_key not in valid_input_names:
            raise AssertionError(
                "limit_key '%s' refers to a non existent "
                "input member for operation: %s, valid keys: "
                "%s"
                % (
                    limit_key,
                    operation_name,
                    ', '.join(list(valid_input_names)),
                )
            )


def _validate_output_keys_match(operation_name, page_config, service_model):
    # NOTE: The original version of this function from translate.py had logic
    # to ensure that the entire set of output_members was accounted for in the
    # union of 'result_key', 'output_token', 'more_results', and
    # 'non_aggregate_keys'.
    # There's enough state drift (especially with non_aggregate_keys) that
    # this is no longer a realistic thing to check.  Someone would have to
    # backport the missing keys to all the paginators.
    output_shape = service_model.operation_model(operation_name).output_shape
    output_members = set(output_shape.members)
    for key_name, output_key in _get_all_page_output_keys(page_config):
        if _looks_like_jmespath(output_key):
            _validate_jmespath_compiles(output_key)
        else:
            if output_key not in output_members:
                raise AssertionError(
                    "Pagination key '%s' refers to an output "
                    "member that does not exist: %s" % (key_name, output_key)
                )
            output_members.remove(output_key)

    for member in list(output_members):
        key = "{}.{}.{}".format(
            service_model.service_name, operation_name, member
        )
        if key in KNOWN_EXTRA_OUTPUT_KEYS:
            output_members.remove(member)

    if output_members:
        raise AssertionError(
            "There are member names in the output shape of "
            "%s that are not accounted for in the pagination "
            "config for service %s: %s"
            % (
                operation_name,
                service_model.service_name,
                ', '.join(output_members),
            )
        )


def _looks_like_jmespath(expression):
    if all(ch in MEMBER_NAME_CHARS for ch in expression):
        return False
    return True


def _validate_jmespath_compiles(expression):
    try:
        jmespath.compile(expression)
    except JMESPathError as e:
        raise AssertionError(
            "Invalid JMESPath expression used "
            "in pagination config: %s\nerror: %s" % (expression, e)
        )


def _get_all_page_output_keys(page_config):
    for key in _get_list_value(page_config, 'result_key'):
        yield 'result_key', key
    for key in _get_list_value(page_config, 'output_token'):
        yield 'output_token', key
    if 'more_results' in page_config:
        yield 'more_results', page_config['more_results']
    for key in page_config.get('non_aggregate_keys', []):
        yield 'non_aggregate_keys', key


def _get_list_value(page_config, key):
    # Some pagination config values can be a scalar value or a list of scalars.
    # This function will always return a list of scalar values, converting as
    # necessary.
    value = page_config[key]
    if not isinstance(value, list):
        value = [value]
    return value



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import pytest

from botocore.client import ClientEndpointBridge
from botocore.exceptions import NoRegionError
from tests import BaseSessionTest, ClientHTTPStubber, mock

# NOTE: sqs endpoint updated to be the CN in the SSL cert because
# a bug in python2.6 prevents subjectAltNames from being parsed
# and subsequently being used in cert validation.
# Same thing is needed for rds.
KNOWN_REGIONS = {
    'ap-northeast-1': {
        'apigateway': 'apigateway.ap-northeast-1.amazonaws.com',
        'appstream': 'appstream.ap-northeast-1.amazonaws.com',
        'autoscaling': 'autoscaling.ap-northeast-1.amazonaws.com',
        'cloudformation': 'cloudformation.ap-northeast-1.amazonaws.com',
        'cloudhsm': 'cloudhsm.ap-northeast-1.amazonaws.com',
        'cloudsearch': 'cloudsearch.ap-northeast-1.amazonaws.com',
        'cloudtrail': 'cloudtrail.ap-northeast-1.amazonaws.com',
        'codedeploy': 'codedeploy.ap-northeast-1.amazonaws.com',
        'cognito-identity': 'cognito-identity.ap-northeast-1.amazonaws.com',
        'cognito-sync': 'cognito-sync.ap-northeast-1.amazonaws.com',
        'config': 'config.ap-northeast-1.amazonaws.com',
        'datapipeline': 'datapipeline.ap-northeast-1.amazonaws.com',
        'directconnect': 'directconnect.ap-northeast-1.amazonaws.com',
        'ds': 'ds.ap-northeast-1.amazonaws.com',
        'dynamodb': 'dynamodb.ap-northeast-1.amazonaws.com',
        'ec2': 'ec2.ap-northeast-1.amazonaws.com',
        'ecs': 'ecs.ap-northeast-1.amazonaws.com',
        'elasticache': 'elasticache.ap-northeast-1.amazonaws.com',
        'elasticbeanstalk': 'elasticbeanstalk.ap-northeast-1.amazonaws.com',
        'elasticloadbalancing': 'elasticloadbalancing.ap-northeast-1.amazonaws.com',
        'elasticmapreduce': 'elasticmapreduce.ap-northeast-1.amazonaws.com',
        'elastictranscoder': 'elastictranscoder.ap-northeast-1.amazonaws.com',
        'glacier': 'glacier.ap-northeast-1.amazonaws.com',
        'iot': 'iot.ap-northeast-1.amazonaws.com',
        'kinesis': 'kinesis.ap-northeast-1.amazonaws.com',
        'kms': 'kms.ap-northeast-1.amazonaws.com',
        'lambda': 'lambda.ap-northeast-1.amazonaws.com',
        'logs': 'logs.ap-northeast-1.amazonaws.com',
        'monitoring': 'monitoring.ap-northeast-1.amazonaws.com',
        'rds': 'rds.ap-northeast-1.amazonaws.com',
        'redshift': 'redshift.ap-northeast-1.amazonaws.com',
        's3': 's3.ap-northeast-1.amazonaws.com',
        'sdb': 'sdb.ap-northeast-1.amazonaws.com',
        'sns': 'sns.ap-northeast-1.amazonaws.com',
        'sqs': 'sqs.ap-northeast-1.amazonaws.com',
        'storagegateway': 'storagegateway.ap-northeast-1.amazonaws.com',
        'streams.dynamodb': 'streams.dynamodb.ap-northeast-1.amazonaws.com',
        'sts': 'sts.ap-northeast-1.amazonaws.com',
        'swf': 'swf.ap-northeast-1.amazonaws.com',
        'workspaces': 'workspaces.ap-northeast-1.amazonaws.com',
    },
    'ap-southeast-1': {
        'autoscaling': 'autoscaling.ap-southeast-1.amazonaws.com',
        'cloudformation': 'cloudformation.ap-southeast-1.amazonaws.com',
        'cloudhsm': 'cloudhsm.ap-southeast-1.amazonaws.com',
        'cloudsearch': 'cloudsearch.ap-southeast-1.amazonaws.com',
        'cloudtrail': 'cloudtrail.ap-southeast-1.amazonaws.com',
        'config': 'config.ap-southeast-1.amazonaws.com',
        'directconnect': 'directconnect.ap-southeast-1.amazonaws.com',
        'ds': 'ds.ap-southeast-1.amazonaws.com',
        'dynamodb': 'dynamodb.ap-southeast-1.amazonaws.com',
        'ec2': 'ec2.ap-southeast-1.amazonaws.com',
        'elasticache': 'elasticache.ap-southeast-1.amazonaws.com',
        'elasticbeanstalk': 'elasticbeanstalk.ap-southeast-1.amazonaws.com',
        'elasticloadbalancing': 'elasticloadbalancing.ap-southeast-1.amazonaws.com',
        'elasticmapreduce': 'elasticmapreduce.ap-southeast-1.amazonaws.com',
        'elastictranscoder': 'elastictranscoder.ap-southeast-1.amazonaws.com',
        'kinesis': 'kinesis.ap-southeast-1.amazonaws.com',
        'kms': 'kms.ap-southeast-1.amazonaws.com',
        'logs': 'logs.ap-southeast-1.amazonaws.com',
        'monitoring': 'monitoring.ap-southeast-1.amazonaws.com',
        'rds': 'rds.ap-southeast-1.amazonaws.com',
        'redshift': 'redshift.ap-southeast-1.amazonaws.com',
        's3': 's3.ap-southeast-1.amazonaws.com',
        'sdb': 'sdb.ap-southeast-1.amazonaws.com',
        'sns': 'sns.ap-southeast-1.amazonaws.com',
        'sqs': 'sqs.ap-southeast-1.amazonaws.com',
        'storagegateway': 'storagegateway.ap-southeast-1.amazonaws.com',
        'streams.dynamodb': 'streams.dynamodb.ap-southeast-1.amazonaws.com',
        'sts': 'sts.ap-southeast-1.amazonaws.com',
        'swf': 'swf.ap-southeast-1.amazonaws.com',
        'workspaces': 'workspaces.ap-southeast-1.amazonaws.com',
    },
    'ap-southeast-2': {
        'autoscaling': 'autoscaling.ap-southeast-2.amazonaws.com',
        'cloudformation': 'cloudformation.ap-southeast-2.amazonaws.com',
        'cloudhsm': 'cloudhsm.ap-southeast-2.amazonaws.com',
        'cloudsearch': 'cloudsearch.ap-southeast-2.amazonaws.com',
        'cloudtrail': 'cloudtrail.ap-southeast-2.amazonaws.com',
        'codedeploy': 'codedeploy.ap-southeast-2.amazonaws.com',
        'config': 'config.ap-southeast-2.amazonaws.com',
        'datapipeline': 'datapipeline.ap-southeast-2.amazonaws.com',
        'directconnect': 'directconnect.ap-southeast-2.amazonaws.com',
        'ds': 'ds.ap-southeast-2.amazonaws.com',
        'dynamodb': 'dynamodb.ap-southeast-2.amazonaws.com',
        'ec2': 'ec2.ap-southeast-2.amazonaws.com',
        'ecs': 'ecs.ap-southeast-2.amazonaws.com',
        'elasticache': 'elasticache.ap-southeast-2.amazonaws.com',
        'elasticbeanstalk': 'elasticbeanstalk.ap-southeast-2.amazonaws.com',
        'elasticloadbalancing': 'elasticloadbalancing.ap-southeast-2.amazonaws.com',
        'elasticmapreduce': 'elasticmapreduce.ap-southeast-2.amazonaws.com',
        'glacier': 'glacier.ap-southeast-2.amazonaws.com',
        'kinesis': 'kinesis.ap-southeast-2.amazonaws.com',
        'kms': 'kms.ap-southeast-2.amazonaws.com',
        'logs': 'logs.ap-southeast-2.amazonaws.com',
        'monitoring': 'monitoring.ap-southeast-2.amazonaws.com',
        'rds': 'rds.ap-southeast-2.amazonaws.com',
        'redshift': 'redshift.ap-southeast-2.amazonaws.com',
        's3': 's3.ap-southeast-2.amazonaws.com',
        'sdb': 'sdb.ap-southeast-2.amazonaws.com',
        'sns': 'sns.ap-southeast-2.amazonaws.com',
        'sqs': 'sqs.ap-southeast-2.amazonaws.com',
        'storagegateway': 'storagegateway.ap-southeast-2.amazonaws.com',
        'streams.dynamodb': 'streams.dynamodb.ap-southeast-2.amazonaws.com',
        'sts': 'sts.ap-southeast-2.amazonaws.com',
        'swf': 'swf.ap-southeast-2.amazonaws.com',
        'workspaces': 'workspaces.ap-southeast-2.amazonaws.com',
    },
    'aws-us-gov-global': {'iam': 'iam.us-gov.amazonaws.com'},
    'cn-north-1': {
        'autoscaling': 'autoscaling.cn-north-1.amazonaws.com.cn',
        'cloudformation': 'cloudformation.cn-north-1.amazonaws.com.cn',
        'cloudtrail': 'cloudtrail.cn-north-1.amazonaws.com.cn',
        'directconnect': 'directconnect.cn-north-1.amazonaws.com.cn',
        'dynamodb': 'dynamodb.cn-north-1.amazonaws.com.cn',
        'ec2': 'ec2.cn-north-1.amazonaws.com.cn',
        'elasticache': 'elasticache.cn-north-1.amazonaws.com.cn',
        'elasticbeanstalk': 'elasticbeanstalk.cn-north-1.amazonaws.com.cn',
        'elasticloadbalancing': 'elasticloadbalancing.cn-north-1.amazonaws.com.cn',
        'elasticmapreduce': 'elasticmapreduce.cn-north-1.amazonaws.com.cn',
        'glacier': 'glacier.cn-north-1.amazonaws.com.cn',
        'iam': 'iam.cn-north-1.amazonaws.com.cn',
        'kinesis': 'kinesis.cn-north-1.amazonaws.com.cn',
        'monitoring': 'monitoring.cn-north-1.amazonaws.com.cn',
        'rds': 'rds.cn-north-1.amazonaws.com.cn',
        's3': 's3.cn-north-1.amazonaws.com.cn',
        'sns': 'sns.cn-north-1.amazonaws.com.cn',
        'sqs': 'sqs.cn-north-1.amazonaws.com.cn',
        'storagegateway': 'storagegateway.cn-north-1.amazonaws.com.cn',
        'streams.dynamodb': 'streams.dynamodb.cn-north-1.amazonaws.com.cn',
        'sts': 'sts.cn-north-1.amazonaws.com.cn',
        'swf': 'swf.cn-north-1.amazonaws.com.cn',
    },
    'eu-central-1': {
        'autoscaling': 'autoscaling.eu-central-1.amazonaws.com',
        'cloudformation': 'cloudformation.eu-central-1.amazonaws.com',
        'cloudhsm': 'cloudhsm.eu-central-1.amazonaws.com',
        'cloudsearch': 'cloudsearch.eu-central-1.amazonaws.com',
        'cloudtrail': 'cloudtrail.eu-central-1.amazonaws.com',
        'codedeploy': 'codedeploy.eu-central-1.amazonaws.com',
        'config': 'config.eu-central-1.amazonaws.com',
        'directconnect': 'directconnect.eu-central-1.amazonaws.com',
        'dynamodb': 'dynamodb.eu-central-1.amazonaws.com',
        'ec2': 'ec2.eu-central-1.amazonaws.com',
        'elasticache': 'elasticache.eu-central-1.amazonaws.com',
        'elasticbeanstalk': 'elasticbeanstalk.eu-central-1.amazonaws.com',
        'elasticloadbalancing': 'elasticloadbalancing.eu-central-1.amazonaws.com',
        'elasticmapreduce': 'elasticmapreduce.eu-central-1.amazonaws.com',
        'glacier': 'glacier.eu-central-1.amazonaws.com',
        'kinesis': 'kinesis.eu-central-1.amazonaws.com',
        'kms': 'kms.eu-central-1.amazonaws.com',
        'logs': 'logs.eu-central-1.amazonaws.com',
        'monitoring': 'monitoring.eu-central-1.amazonaws.com',
        'rds': 'rds.eu-central-1.amazonaws.com',
        'redshift': 'redshift.eu-central-1.amazonaws.com',
        's3': 's3.eu-central-1.amazonaws.com',
        'sns': 'sns.eu-central-1.amazonaws.com',
        'sqs': 'sqs.eu-central-1.amazonaws.com',
        'storagegateway': 'storagegateway.eu-central-1.amazonaws.com',
        'streams.dynamodb': 'streams.dynamodb.eu-central-1.amazonaws.com',
        'sts': 'sts.eu-central-1.amazonaws.com',
        'swf': 'swf.eu-central-1.amazonaws.com',
    },
    'eu-west-1': {
        'apigateway': 'apigateway.eu-west-1.amazonaws.com',
        'autoscaling': 'autoscaling.eu-west-1.amazonaws.com',
        'cloudformation': 'cloudformation.eu-west-1.amazonaws.com',
        'cloudhsm': 'cloudhsm.eu-west-1.amazonaws.com',
        'cloudsearch': 'cloudsearch.eu-west-1.amazonaws.com',
        'cloudtrail': 'cloudtrail.eu-west-1.amazonaws.com',
        'codedeploy': 'codedeploy.eu-west-1.amazonaws.com',
        'cognito-identity': 'cognito-identity.eu-west-1.amazonaws.com',
        'cognito-sync': 'cognito-sync.eu-west-1.amazonaws.com',
        'config': 'config.eu-west-1.amazonaws.com',
        'datapipeline': 'datapipeline.eu-west-1.amazonaws.com',
        'directconnect': 'directconnect.eu-west-1.amazonaws.com',
        'ds': 'ds.eu-west-1.amazonaws.com',
        'dynamodb': 'dynamodb.eu-west-1.amazonaws.com',
        'ec2': 'ec2.eu-west-1.amazonaws.com',
        'ecs': 'ecs.eu-west-1.amazonaws.com',
        'elasticache': 'elasticache.eu-west-1.amazonaws.com',
        'elasticbeanstalk': 'elasticbeanstalk.eu-west-1.amazonaws.com',
        'elasticloadbalancing': 'elasticloadbalancing.eu-west-1.amazonaws.com',
        'elasticmapreduce': 'elasticmapreduce.eu-west-1.amazonaws.com',
        'elastictranscoder': 'elastictranscoder.eu-west-1.amazonaws.com',
        'email': 'email.eu-west-1.amazonaws.com',
        'glacier': 'glacier.eu-west-1.amazonaws.com',
        'iot': 'iot.eu-west-1.amazonaws.com',
        'kinesis': 'kinesis.eu-west-1.amazonaws.com',
        'kms': 'kms.eu-west-1.amazonaws.com',
        'lambda': 'lambda.eu-west-1.amazonaws.com',
        'logs': 'logs.eu-west-1.amazonaws.com',
        'machinelearning': 'machinelearning.eu-west-1.amazonaws.com',
        'monitoring': 'monitoring.eu-west-1.amazonaws.com',
        'rds': 'rds.eu-west-1.amazonaws.com',
        'redshift': 'redshift.eu-west-1.amazonaws.com',
        's3': 's3.eu-west-1.amazonaws.com',
        'sdb': 'sdb.eu-west-1.amazonaws.com',
        'sns': 'sns.eu-west-1.amazonaws.com',
        'sqs': 'sqs.eu-west-1.amazonaws.com',
        'ssm': 'ssm.eu-west-1.amazonaws.com',
        'storagegateway': 'storagegateway.eu-west-1.amazonaws.com',
        'streams.dynamodb': 'streams.dynamodb.eu-west-1.amazonaws.com',
        'sts': 'sts.eu-west-1.amazonaws.com',
        'swf': 'swf.eu-west-1.amazonaws.com',
        'workspaces': 'workspaces.eu-west-1.amazonaws.com',
    },
    'fips-us-gov-west-1': {'s3': 's3-fips.us-gov-west-1.amazonaws.com'},
    's3-external-1': {'s3': 's3-external-1.amazonaws.com'},
    'sa-east-1': {
        'autoscaling': 'autoscaling.sa-east-1.amazonaws.com',
        'cloudformation': 'cloudformation.sa-east-1.amazonaws.com',
        'cloudsearch': 'cloudsearch.sa-east-1.amazonaws.com',
        'cloudtrail': 'cloudtrail.sa-east-1.amazonaws.com',
        'config': 'config.sa-east-1.amazonaws.com',
        'directconnect': 'directconnect.sa-east-1.amazonaws.com',
        'dynamodb': 'dynamodb.sa-east-1.amazonaws.com',
        'ec2': 'ec2.sa-east-1.amazonaws.com',
        'elasticache': 'elasticache.sa-east-1.amazonaws.com',
        'elasticbeanstalk': 'elasticbeanstalk.sa-east-1.amazonaws.com',
        'elasticloadbalancing': 'elasticloadbalancing.sa-east-1.amazonaws.com',
        'elasticmapreduce': 'elasticmapreduce.sa-east-1.amazonaws.com',
        'kms': 'kms.sa-east-1.amazonaws.com',
        'monitoring': 'monitoring.sa-east-1.amazonaws.com',
        'rds': 'rds.sa-east-1.amazonaws.com',
        's3': 's3.sa-east-1.amazonaws.com',
        'sdb': 'sdb.sa-east-1.amazonaws.com',
        'sns': 'sns.sa-east-1.amazonaws.com',
        'sqs': 'sqs.sa-east-1.amazonaws.com',
        'storagegateway': 'storagegateway.sa-east-1.amazonaws.com',
        'streams.dynamodb': 'streams.dynamodb.sa-east-1.amazonaws.com',
        'sts': 'sts.sa-east-1.amazonaws.com',
        'swf': 'swf.sa-east-1.amazonaws.com',
    },
    'us-east-1': {
        'apigateway': 'apigateway.us-east-1.amazonaws.com',
        'appstream': 'appstream.us-east-1.amazonaws.com',
        'autoscaling': 'autoscaling.us-east-1.amazonaws.com',
        'cloudformation': 'cloudformation.us-east-1.amazonaws.com',
        'cloudfront': 'cloudfront.amazonaws.com',
        'cloudhsm': 'cloudhsm.us-east-1.amazonaws.com',
        'cloudsearch': 'cloudsearch.us-east-1.amazonaws.com',
        'cloudtrail': 'cloudtrail.us-east-1.amazonaws.com',
        'codecommit': 'codecommit.us-east-1.amazonaws.com',
        'codedeploy': 'codedeploy.us-east-1.amazonaws.com',
        'codepipeline': 'codepipeline.us-east-1.amazonaws.com',
        'cognito-identity': 'cognito-identity.us-east-1.amazonaws.com',
        'cognito-sync': 'cognito-sync.us-east-1.amazonaws.com',
        'config': 'config.us-east-1.amazonaws.com',
        'datapipeline': 'datapipeline.us-east-1.amazonaws.com',
        'directconnect': 'directconnect.us-east-1.amazonaws.com',
        'ds': 'ds.us-east-1.amazonaws.com',
        'dynamodb': 'dynamodb.us-east-1.amazonaws.com',
        'ec2': 'ec2.us-east-1.amazonaws.com',
        'ecs': 'ecs.us-east-1.amazonaws.com',
        'elasticache': 'elasticache.us-east-1.amazonaws.com',
        'elasticbeanstalk': 'elasticbeanstalk.us-east-1.amazonaws.com',
        'elasticloadbalancing': 'elasticloadbalancing.us-east-1.amazonaws.com',
        'elasticmapreduce': 'elasticmapreduce.us-east-1.amazonaws.com',
        'elastictranscoder': 'elastictranscoder.us-east-1.amazonaws.com',
        'email': 'email.us-east-1.amazonaws.com',
        'glacier': 'glacier.us-east-1.amazonaws.com',
        'iam': 'iam.amazonaws.com',
        'importexport': 'importexport.amazonaws.com',
        'iot': 'iot.us-east-1.amazonaws.com',
        'kinesis': 'kinesis.us-east-1.amazonaws.com',
        'kms': 'kms.us-east-1.amazonaws.com',
        'lambda': 'lambda.us-east-1.amazonaws.com',
        'logs': 'logs.us-east-1.amazonaws.com',
        'machinelearning': 'machinelearning.us-east-1.amazonaws.com',
        'mobileanalytics': 'mobileanalytics.us-east-1.amazonaws.com',
        'monitoring': 'monitoring.us-east-1.amazonaws.com',
        'opsworks': 'opsworks.us-east-1.amazonaws.com',
        'rds': 'rds.us-east-1.amazonaws.com',
        'redshift': 'redshift.us-east-1.amazonaws.com',
        'route53': 'route53.amazonaws.com',
        'route53domains': 'route53domains.us-east-1.amazonaws.com',
        's3': 's3.us-east-1.amazonaws.com',
        'sdb': 'sdb.amazonaws.com',
        'sns': 'sns.us-east-1.amazonaws.com',
        'sqs': 'sqs.us-east-1.amazonaws.com',
        'ssm': 'ssm.us-east-1.amazonaws.com',
        'storagegateway': 'storagegateway.us-east-1.amazonaws.com',
        'streams.dynamodb': 'streams.dynamodb.us-east-1.amazonaws.com',
        'sts': 'sts.us-east-1.amazonaws.com',
        'support': 'support.us-east-1.amazonaws.com',
        'swf': 'swf.us-east-1.amazonaws.com',
        'workspaces': 'workspaces.us-east-1.amazonaws.com',
        'waf': 'waf.amazonaws.com',
    },
    'us-gov-west-1': {
        'autoscaling': 'autoscaling.us-gov-west-1.amazonaws.com',
        'cloudformation': 'cloudformation.us-gov-west-1.amazonaws.com',
        'cloudhsm': 'cloudhsm.us-gov-west-1.amazonaws.com',
        'cloudtrail': 'cloudtrail.us-gov-west-1.amazonaws.com',
        'dynamodb': 'dynamodb.us-gov-west-1.amazonaws.com',
        'ec2': 'ec2.us-gov-west-1.amazonaws.com',
        'elasticache': 'elasticache.us-gov-west-1.amazonaws.com',
        'elasticloadbalancing': 'elasticloadbalancing.us-gov-west-1.amazonaws.com',
        'elasticmapreduce': 'elasticmapreduce.us-gov-west-1.amazonaws.com',
        'glacier': 'glacier.us-gov-west-1.amazonaws.com',
        'iam': 'iam.us-gov.amazonaws.com',
        'kms': 'kms.us-gov-west-1.amazonaws.com',
        'monitoring': 'monitoring.us-gov-west-1.amazonaws.com',
        'rds': 'rds.us-gov-west-1.amazonaws.com',
        'redshift': 'redshift.us-gov-west-1.amazonaws.com',
        's3': 's3.us-gov-west-1.amazonaws.com',
        'sns': 'sns.us-gov-west-1.amazonaws.com',
        'sqs': 'sqs.us-gov-west-1.amazonaws.com',
        'sts': 'sts.us-gov-west-1.amazonaws.com',
        'swf': 'swf.us-gov-west-1.amazonaws.com',
    },
    'us-west-1': {
        'autoscaling': 'autoscaling.us-west-1.amazonaws.com',
        'cloudformation': 'cloudformation.us-west-1.amazonaws.com',
        'cloudsearch': 'cloudsearch.us-west-1.amazonaws.com',
        'cloudtrail': 'cloudtrail.us-west-1.amazonaws.com',
        'config': 'config.us-west-1.amazonaws.com',
        'directconnect': 'directconnect.us-west-1.amazonaws.com',
        'dynamodb': 'dynamodb.us-west-1.amazonaws.com',
        'ec2': 'ec2.us-west-1.amazonaws.com',
        'ecs': 'ecs.us-west-1.amazonaws.com',
        'elasticache': 'elasticache.us-west-1.amazonaws.com',
        'elasticbeanstalk': 'elasticbeanstalk.us-west-1.amazonaws.com',
        'elasticloadbalancing': 'elasticloadbalancing.us-west-1.amazonaws.com',
        'elasticmapreduce': 'elasticmapreduce.us-west-1.amazonaws.com',
        'elastictranscoder': 'elastictranscoder.us-west-1.amazonaws.com',
        'glacier': 'glacier.us-west-1.amazonaws.com',
        'kinesis': 'kinesis.us-west-1.amazonaws.com',
        'kms': 'kms.us-west-1.amazonaws.com',
        'logs': 'logs.us-west-1.amazonaws.com',
        'monitoring': 'monitoring.us-west-1.amazonaws.com',
        'rds': 'rds.us-west-1.amazonaws.com',
        's3': 's3.us-west-1.amazonaws.com',
        'sdb': 'sdb.us-west-1.amazonaws.com',
        'sns': 'sns.us-west-1.amazonaws.com',
        'sqs': 'sqs.us-west-1.amazonaws.com',
        'storagegateway': 'storagegateway.us-west-1.amazonaws.com',
        'streams.dynamodb': 'streams.dynamodb.us-west-1.amazonaws.com',
        'sts': 'sts.us-west-1.amazonaws.com',
        'swf': 'swf.us-west-1.amazonaws.com',
    },
    'us-west-2': {
        'apigateway': 'apigateway.us-west-2.amazonaws.com',
        'autoscaling': 'autoscaling.us-west-2.amazonaws.com',
        'cloudformation': 'cloudformation.us-west-2.amazonaws.com',
        'cloudhsm': 'cloudhsm.us-west-2.amazonaws.com',
        'cloudsearch': 'cloudsearch.us-west-2.amazonaws.com',
        'cloudtrail': 'cloudtrail.us-west-2.amazonaws.com',
        'codedeploy': 'codedeploy.us-west-2.amazonaws.com',
        'codepipeline': 'codepipeline.us-west-2.amazonaws.com',
        'config': 'config.us-west-2.amazonaws.com',
        'datapipeline': 'datapipeline.us-west-2.amazonaws.com',
        'devicefarm': 'devicefarm.us-west-2.amazonaws.com',
        'directconnect': 'directconnect.us-west-2.amazonaws.com',
        'ds': 'ds.us-west-2.amazonaws.com',
        'dynamodb': 'dynamodb.us-west-2.amazonaws.com',
        'ec2': 'ec2.us-west-2.amazonaws.com',
        'ecs': 'ecs.us-west-2.amazonaws.com',
        'elasticache': 'elasticache.us-west-2.amazonaws.com',
        'elasticbeanstalk': 'elasticbeanstalk.us-west-2.amazonaws.com',
        'elasticfilesystem': 'elasticfilesystem.us-west-2.amazonaws.com',
        'elasticloadbalancing': 'elasticloadbalancing.us-west-2.amazonaws.com',
        'elasticmapreduce': 'elasticmapreduce.us-west-2.amazonaws.com',
        'elastictranscoder': 'elastictranscoder.us-west-2.amazonaws.com',
        'email': 'email.us-west-2.amazonaws.com',
        'glacier': 'glacier.us-west-2.amazonaws.com',
        'iot': 'iot.us-west-2.amazonaws.com',
        'kinesis': 'kinesis.us-west-2.amazonaws.com',
        'kms': 'kms.us-west-2.amazonaws.com',
        'lambda': 'lambda.us-west-2.amazonaws.com',
        'logs': 'logs.us-west-2.amazonaws.com',
        'monitoring': 'monitoring.us-west-2.amazonaws.com',
        'rds': 'rds.us-west-2.amazonaws.com',
        'redshift': 'redshift.us-west-2.amazonaws.com',
        's3': 's3.us-west-2.amazonaws.com',
        'sdb': 'sdb.us-west-2.amazonaws.com',
        'sns': 'sns.us-west-2.amazonaws.com',
        'sqs': 'sqs.us-west-2.amazonaws.com',
        'ssm': 'ssm.us-west-2.amazonaws.com',
        'storagegateway': 'storagegateway.us-west-2.amazonaws.com',
        'streams.dynamodb': 'streams.dynamodb.us-west-2.amazonaws.com',
        'sts': 'sts.us-west-2.amazonaws.com',
        'swf': 'swf.us-west-2.amazonaws.com',
        'workspaces': 'workspaces.us-west-2.amazonaws.com',
    },
}


# Lists the services in the aws partition that do not require a region
# when resolving an endpoint because these services have partitionWide
# endpoints.
KNOWN_AWS_PARTITION_WIDE = {
    'importexport': 'https://importexport.amazonaws.com',
    'cloudfront': 'https://cloudfront.amazonaws.com',
    'waf': 'https://waf.amazonaws.com',
    'route53': 'https://route53.amazonaws.com',
    's3': 'https://s3.amazonaws.com',
    'sts': 'https://sts.amazonaws.com',
    'iam': 'https://iam.amazonaws.com',
}


def _known_endpoints_by_region():
    for region_name, service_dict in KNOWN_REGIONS.items():
        for service_name, endpoint in service_dict.items():
            yield service_name, region_name, endpoint


@pytest.mark.parametrize(
    "service_name, region_name, expected_endpoint",
    _known_endpoints_by_region(),
)
def test_single_service_region_endpoint(
    patched_session, service_name, region_name, expected_endpoint
):
    # Verify the actual values from the partition files.  While
    # TestEndpointHeuristics verified the generic functionality given any
    # endpoints file, this test actually verifies the partition data against a
    # fixed list of known endpoints.  This list doesn't need to be kept 100% up
    # to date, but serves as a basis for regressions as the endpoint data
    # logic evolves.
    resolver = patched_session._get_internal_component('endpoint_resolver')
    bridge = ClientEndpointBridge(resolver, None, None)
    result = bridge.resolve(service_name, region_name)
    expected = 'https://%s' % expected_endpoint
    assert result['endpoint_url'] == expected


# Ensure that all S3 regions use s3v4 instead of v4
def test_all_s3_endpoints_have_s3v4(patched_session):
    session = patched_session
    partitions = session.get_available_partitions()
    resolver = session._get_internal_component('endpoint_resolver')
    for partition_name in partitions:
        for endpoint in session.get_available_regions('s3', partition_name):
            resolved = resolver.construct_endpoint('s3', endpoint)
            assert 's3v4' in resolved['signatureVersions']
            assert 'v4' not in resolved['signatureVersions']


@pytest.mark.parametrize(
    "service_name, expected_endpoint", KNOWN_AWS_PARTITION_WIDE.items()
)
def test_single_service_partition_endpoint(
    patched_session, service_name, expected_endpoint
):
    resolver = patched_session._get_internal_component('endpoint_resolver')
    bridge = ClientEndpointBridge(resolver)
    result = bridge.resolve(service_name)
    assert result['endpoint_url'] == expected_endpoint


def test_non_partition_endpoint_requires_region(patched_session):
    resolver = patched_session._get_internal_component('endpoint_resolver')
    with pytest.raises(NoRegionError):
        resolver.construct_endpoint('ec2')


class TestEndpointResolution(BaseSessionTest):
    def setUp(self):
        super().setUp()
        self.xml_response = (
            b'<?xml version="1.0" encoding="UTF-8"?>\n\n'
            b'<ListRolesResponse '
            b'xmlns="https://iam.amazonaws.com/doc/2010-05-08/">\n'
            b'<ListRolesResult>\n</ListRolesResult>'
            b'</ListRolesResponse>'
        )

    def create_stubbed_client(self, service_name, region_name, **kwargs):
        client = self.session.create_client(
            service_name, region_name, **kwargs
        )
        http_stubber = ClientHTTPStubber(client)
        http_stubber.start()
        return client, http_stubber

    def test_regionalized_client_endpoint_resolution(self):
        client, stubber = self.create_stubbed_client('s3', 'us-east-2')
        stubber.add_response()
        client.list_buckets()
        self.assertEqual(
            stubber.requests[0].url, 'https://s3.us-east-2.amazonaws.com/'
        )

    def test_regionalized_client_with_unknown_region(self):
        client, stubber = self.create_stubbed_client('s3', 'not-real')
        stubber.add_response()
        client.list_buckets()
        # Validate we don't fall back to partition endpoint for
        # regionalized services.
        self.assertEqual(
            stubber.requests[0].url, 'https://s3.not-real.amazonaws.com/'
        )

    def test_unregionalized_client_endpoint_resolution(self):
        client, stubber = self.create_stubbed_client('iam', 'us-west-2')
        stubber.add_response(body=self.xml_response)
        client.list_roles()
        self.assertTrue(
            stubber.requests[0].url.startswith('https://iam.amazonaws.com/')
        )

    def test_unregionalized_client_with_unknown_region(self):
        client, stubber = self.create_stubbed_client('iam', 'not-real')
        stubber.add_response(body=self.xml_response)
        client.list_roles()
        self.assertTrue(
            stubber.requests[0].url.startswith('https://iam.amazonaws.com/')
        )


@pytest.mark.parametrize("is_builtin", [True, False])
def test_endpoint_resolver_knows_its_datasource(patched_session, is_builtin):
    # The information whether or not the endpoints.json file was loaded from
    # the builtin data directory or not should be passed from Loader to
    # EndpointResolver.
    session = patched_session
    loader = session.get_component('data_loader')
    with mock.patch.object(loader, 'is_builtin_path', return_value=is_builtin):
        resolver = session._get_internal_component('endpoint_resolver')
        assert resolver.uses_builtin_data == is_builtin



import json

from behave import then, when

from botocore import xform_name
from botocore.exceptions import ClientError


def _params_from_table(table):
    # Unfortunately the way we're using table is not quite how
    # behave expects tables to be used:
    # They expect:
    #
    #     | name      | department  |
    #     | Barry     | foo         |
    #     | Pudey     | bar         |
    #     | Two-Lumps | bar         |
    #
    # Where the first row are headings that indicate the
    # key name you can use to retrieve row values,
    # e.g row['name'] -> Barry.
    #
    #
    # We just use:
    #      | LaunchConfigurationName | hello, world |
    #      | ImageId                 | ami-12345678 |
    #      | InstanceType            | m1.small     |
    #
    # So we have to grab the headings before iterating over
    # the table rows.
    params = {table.headings[0]: table.headings[1]}
    for row in table:
        params[row[0]] = row[1]
    return params


@when('I call the "{}" API')
def api_call_no_args(context, operation):
    context.response = getattr(context.client, xform_name(operation))()


@when('I call the "{}" API with')
def api_call_with_args(context, operation):
    params = _params_from_table(context.table)
    context.response = getattr(context.client, xform_name(operation))(**params)


@when('I call the "{}" API with JSON')
def api_call_with_json(context, operation):
    params = json.loads(context.text)
    context.response = getattr(context.client, xform_name(operation))(**params)


@when('I attempt to call the "{}" API with')
def api_call_with_error(context, operation):
    params = _params_from_table(context.table)
    try:
        getattr(context.client, xform_name(operation))(**params)
    except ClientError as e:
        context.error_response = e


@when('I attempt to call the "{}" API with JSON')
def api_call_with_json_and_error(context, operation):
    params = json.loads(context.text)
    try:
        getattr(context.client, xform_name(operation))(**params)
    except ClientError as e:
        context.error_response = e


@then('I expect the response error code to be "{}"')
def then_expected_error(context, code):
    assert context.error_response.response['Error']['Code'] == code


@then('the value at "{}" should be a list')
def then_expected_type_is_list(context, expression):
    # In botocore, if there are no values with an element,
    # it will not appear in the response dict, so it's actually
    # ok if the element does not exist (and is not a list).
    # If an exception happened the test will have already failed,
    # which makes this step a noop.  We'll just verify
    # the response is a dict to ensure it made it through
    # our response parser properly.
    if not isinstance(context.response, dict):
        raise AssertionError("Response is not a dict: %s" % context.response)


@then('the response should contain a "{}"')
def then_should_contain_key(context, key):
    # See then_expected_type_is_a_list for more background info.
    # We really just care that the request succeeded for these
    # smoke tests.
    if not isinstance(context.response, dict):
        raise AssertionError("Response is not a dict: %s" % context.response)


@then('I expect the response error to contain a message')
def then_error_has_message(context):
    if 'Message' not in context.error_response.response['Error']:
        raise AssertionError(
            "Message key missing from error response: %s"
            % context.error_response.response
        )



# Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import platform

import pytest

import botocore.useragent
from botocore import __version__ as botocore_version
from botocore.config import Config
from botocore.useragent import (
    UserAgentString,
    sanitize_user_agent_string_component,
)
from tests import mock

from .. import requires_crt


# Returns a list of unmodified User-Agent components.
def unmodified_components(components):
    return components


@pytest.mark.parametrize(
    'raw_str, allow_hash, expected_str',
    [
        ('foo', False, 'foo'),
        ('foo', True, 'foo'),
        ('ExampleFramework (1.2.3)', False, 'ExampleFramework--1.2.3-'),
        ('foo#1.2.3', False, 'foo-1.2.3'),
        ('foo#1.2.3', True, 'foo#1.2.3'),
        ('', False, ''),
        ('', True, ''),
        ('', False, ''),
        ('#', False, '-'),
        ('#', True, '#'),
        (' ', False, '-'),
        ('  ', False, '--'),
        ('@=[]{ }/\\øß©', True, '------------'),
        (
            'Java_HotSpot_(TM)_64-Bit_Server_VM/25.151-b12',
            True,
            'Java_HotSpot_-TM-_64-Bit_Server_VM-25.151-b12',
        ),
    ],
)
def test_sanitize_ua_string_component(raw_str, allow_hash, expected_str):
    actual_str = sanitize_user_agent_string_component(raw_str, allow_hash)
    assert actual_str == expected_str


@mock.patch.object(
    botocore.useragent, 'modify_components', unmodified_components
)
def test_basic_user_agent_string():
    ua = UserAgentString(
        platform_name='linux',
        platform_version='1.2.3-foo',
        platform_machine='x86_64',
        python_version='3.8.20',
        python_implementation='Dpython',
        execution_env='AWS_Lambda_python3.8',
        crt_version='Unknown',
    ).with_client_config(
        Config(retries={'mode': 'legacy'}, user_agent_appid='fooapp')
    )

    actual = ua.to_string()
    expected = (
        f'Botocore/{botocore_version} '
        'md/awscrt#Unknown '
        'ua/2.0 '
        'os/linux#1.2.3-foo '
        'md/arch#x86_64 '
        'lang/python#3.8.20 '
        'md/pyimpl#Dpython '
        'exec-env/AWS_Lambda_python3.8 '
        'cfg/retry-mode#legacy '
        'app/fooapp'
    )
    assert actual == expected


def test_shared_test_case():
    # This test case is shared across AWS SDKs.

    uas = UserAgentString(
        platform_name="Linux",
        platform_version="5.4.228-131.415.AMZN2.X86_64",
        platform_machine="",
        python_version="4.3.2",
        python_implementation=None,
        execution_env='lambda',
    ).with_client_config(
        Config(user_agent_appid='123456', retries={'mode': 'standard'})
    )
    actual = uas.to_string().split(' ')
    expected_in_exact_order = [
        f"Botocore/{botocore_version}",
        "ua/2.0",
        "os/linux#5.4.228-131.415.AMZN2.X86_64",
        "lang/python#4.3.2",
        "exec-env/lambda",
    ]
    expected_in_any_order = [
        "cfg/retry-mode#standard",
        "app/123456",
    ]
    for el in [*expected_in_exact_order, *expected_in_any_order]:
        assert el in actual

    indices = [actual.index(el) for el in expected_in_exact_order]
    assert indices == list(sorted(indices)), 'Elements were found out of order'


@mock.patch.object(
    botocore.useragent, 'modify_components', unmodified_components
)
def test_user_agent_string_with_missing_information():
    # Even when collecting information from the environment fails completely,
    # some minimal string should be generated.
    uas = UserAgentString(
        platform_name=None,
        platform_version=None,
        platform_machine=None,
        python_version=None,
        python_implementation=None,
        execution_env=None,
        crt_version=None,
    ).with_client_config(Config())
    actual = uas.to_string()
    assert actual == f'Botocore/{botocore_version} ua/2.0 os/other lang/python'


def test_from_environment(monkeypatch):
    monkeypatch.setenv('AWS_EXECUTION_ENV', 'lambda')
    monkeypatch.setattr(platform, 'system', lambda: 'Linux')
    monkeypatch.setattr(
        platform, 'release', lambda: '5.4.228-131.415.AMZN2.X86_64'
    )
    monkeypatch.setattr(platform, 'python_version', lambda: '4.3.2')
    monkeypatch.setattr(platform, 'python_implementation', lambda: 'Cpython')

    uas = UserAgentString.from_environment()

    assert uas._execution_env == 'lambda'
    assert uas._platform_name == 'Linux'
    assert uas._platform_version == '5.4.228-131.415.AMZN2.X86_64'
    assert uas._python_version == '4.3.2'
    assert uas._python_implementation == 'Cpython'


@requires_crt()
def test_from_environment_can_read_crt_version(monkeypatch):
    import awscrt

    monkeypatch.setattr(awscrt, '__version__', 'a.b.c')
    uas = UserAgentString.from_environment()
    assert uas._crt_version == 'a.b.c'


def test_from_environment_with_most_values_not_available(monkeypatch):
    # Asserts that ``None`` values are properly passed through to the
    # UserAgentString class. There are separate tests to assert that
    # ``UserAgentString.to_string()`` can handle ``None`` values.
    monkeypatch.delenv('AWS_EXECUTION_ENV', raising=False)
    monkeypatch.setattr(platform, 'system', lambda: None)
    monkeypatch.setattr(platform, 'release', lambda: None)
    monkeypatch.setattr(platform, 'python_version', lambda: None)
    monkeypatch.setattr(platform, 'python_implementation', lambda: None)

    uas = UserAgentString.from_environment()

    assert uas._execution_env is None
    assert uas._platform_name is None
    assert uas._platform_version is None
    assert uas._python_version is None
    assert uas._python_implementation is None


def test_from_environment_unknown_platform(monkeypatch):
    monkeypatch.setattr(platform, 'system', lambda: 'FooOS')
    monkeypatch.setattr(platform, 'release', lambda: '0.0.1')
    uas = UserAgentString.from_environment()
    assert ' os/other md/FooOS#0.0.1 ' in uas.to_string()



#!/usr/bin/env python
# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Test runner for the JSON models compliance tests

This is a test runner for all the JSON tests defined in
``tests/unit/protocols/``, including both the input/output tests.

You can use the normal ``python -m pytest tests/unit/test_protocols.py``
to run this test.  In addition, there are several env vars you can use during
development.

Tests are broken down by filename, test suite, testcase.  When a test fails
you'll see the protocol (filename), test suite, and test case number of the
failed test.

::

    Description           : Scalar members (0:0)  <--- (suite_id:test_id)
    Protocol:             : ec2                  <--- test file (ec2.json)
    Given                 : ...
    Response              : ...
    Expected serialization: ...
    Actual serialization  : ...
    Assertion message     : ...

To run tests from only a single file, you can set the
BOTOCORE_TEST env var::

    BOTOCORE_TEST=tests/unit/compliance/input/json.json pytest tests/unit/test_protocols.py

To run a single test suite you can set the BOTOCORE_TEST_ID env var:

    BOTOCORE_TEST=tests/unit/compliance/input/json.json BOTOCORE_TEST_ID=5 \
        pytest tests/unit/test_protocols.py

To run a single test case in a suite (useful when debugging a single test), you
can set the BOTOCORE_TEST_ID env var with the ``suite_id:test_id`` syntax.

    BOTOCORE_TEST_ID=5:1 pytest test/unit/test_protocols.py

"""
import copy
import os
from base64 import b64decode
from calendar import timegm
from enum import Enum

import pytest
from dateutil.tz import tzutc

from botocore.awsrequest import HeadersDict, prepare_request_dict
from botocore.compat import OrderedDict, json, urlsplit
from botocore.eventstream import EventStream
from botocore.model import NoShapeFoundError, OperationModel, ServiceModel
from botocore.parsers import (
    EC2QueryParser,
    JSONParser,
    QueryParser,
    RestJSONParser,
    RestXMLParser,
)
from botocore.serialize import (
    EC2Serializer,
    JSONSerializer,
    QuerySerializer,
    RestJSONSerializer,
    RestXMLSerializer,
)
from botocore.utils import parse_timestamp, percent_encode_sequence

TEST_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'protocols'
)
NOT_SPECIFIED = object()
PROTOCOL_SERIALIZERS = {
    'ec2': EC2Serializer,
    'query': QuerySerializer,
    'json': JSONSerializer,
    'rest-json': RestJSONSerializer,
    'rest-xml': RestXMLSerializer,
}
PROTOCOL_PARSERS = {
    'ec2': EC2QueryParser,
    'query': QueryParser,
    'json': JSONParser,
    'rest-json': RestJSONParser,
    'rest-xml': RestXMLParser,
}
PROTOCOL_TEST_BLACKLIST = ['Idempotency token auto fill']


class TestType(Enum):
    # Tell test runner to ignore this class
    __test__ = False

    INPUT = "input"
    OUTPUT = "output"


def _compliance_tests(test_type=None):
    inp = test_type is None or test_type is TestType.INPUT
    out = test_type is None or test_type is TestType.OUTPUT

    for full_path in _walk_files():
        if full_path.endswith('.json'):
            for model, case, basename in _load_cases(full_path):
                if model.get('description') in PROTOCOL_TEST_BLACKLIST:
                    continue
                if 'params' in case and inp:
                    yield model, case, basename
                elif 'response' in case and out:
                    yield model, case, basename


@pytest.mark.parametrize(
    "json_description, case, basename", _compliance_tests(TestType.INPUT)
)
def test_input_compliance(json_description, case, basename):
    service_description = copy.deepcopy(json_description)
    service_description['operations'] = {
        case.get('name', 'OperationName'): case,
    }
    model = ServiceModel(service_description)
    protocol_type = model.metadata['protocol']
    try:
        protocol_serializer = PROTOCOL_SERIALIZERS[protocol_type]
    except KeyError:
        raise RuntimeError("Unknown protocol: %s" % protocol_type)
    serializer = protocol_serializer()
    serializer.MAP_TYPE = OrderedDict
    operation_model = OperationModel(case['given'], model)
    request = serializer.serialize_to_request(case['params'], operation_model)
    _serialize_request_description(request)
    client_endpoint = service_description.get('clientEndpoint')
    try:
        _assert_request_body_is_bytes(request['body'])
        _assert_requests_equal(request, case['serialized'])
        _assert_endpoints_equal(request, case['serialized'], client_endpoint)
    except AssertionError as e:
        _input_failure_message(protocol_type, case, request, e)


def _assert_request_body_is_bytes(body):
    if not isinstance(body, bytes):
        raise AssertionError(
            "Expected body to be serialized as type "
            "bytes(), instead got: %s" % type(body)
        )


def _assert_endpoints_equal(actual, expected, endpoint):
    if 'host' not in expected:
        return
    prepare_request_dict(actual, endpoint)
    actual_host = urlsplit(actual['url']).netloc
    assert_equal(actual_host, expected['host'], 'Host')


class MockRawResponse:
    def __init__(self, data):
        self._data = b64decode(data)

    def stream(self):
        yield self._data


@pytest.mark.parametrize(
    "json_description, case, basename", _compliance_tests(TestType.OUTPUT)
)
def test_output_compliance(json_description, case, basename):
    service_description = copy.deepcopy(json_description)
    operation_name = case.get('name', 'OperationName')
    service_description['operations'] = {
        operation_name: case,
    }
    case['response']['context'] = {'operation_name': operation_name}
    try:
        model = ServiceModel(service_description)
        operation_model = OperationModel(case['given'], model)
        parser = PROTOCOL_PARSERS[model.metadata['protocol']](
            timestamp_parser=_compliance_timestamp_parser
        )
        # We load the json as utf-8, but the response parser is at the
        # botocore boundary, so it expects to work with bytes.
        body_bytes = case['response']['body'].encode('utf-8')
        case['response']['body'] = body_bytes
        # We need the headers to be case insensitive
        headers = HeadersDict(case['response']['headers'])
        case['response']['headers'] = headers
        # If this is an event stream fake the raw streamed response
        if operation_model.has_event_stream_output:
            case['response']['body'] = MockRawResponse(body_bytes)
        if 'error' in case:
            output_shape = operation_model.output_shape
            parsed = parser.parse(case['response'], output_shape)
            try:
                error_shape = model.shape_for(parsed['Error']['Code'])
            except NoShapeFoundError:
                error_shape = None
            if error_shape is not None:
                error_parse = parser.parse(case['response'], error_shape)
                parsed.update(error_parse)
        else:
            output_shape = operation_model.output_shape
            parsed = parser.parse(case['response'], output_shape)
        parsed = _fixup_parsed_result(parsed)
    except Exception as e:
        msg = (
            "\nFailed to run test  : %s\n"
            "Protocol            : %s\n"
            "Description         : %s (%s:%s)\n"
            % (
                e,
                model.metadata['protocol'],
                case['description'],
                case['suite_id'],
                case['test_id'],
            )
        )
        raise AssertionError(msg)
    try:
        if 'error' in case:
            expected_result = {
                'Error': {
                    'Code': case.get('errorCode', ''),
                    'Message': case.get('errorMessage', ''),
                }
            }
            expected_result.update(case['error'])
        else:
            expected_result = case['result']
        assert_equal(parsed, expected_result, "Body")
    except Exception as e:
        _output_failure_message(
            model.metadata['protocol'], case, parsed, expected_result, e
        )


def _fixup_parsed_result(parsed):
    # This function contains all the transformation we need
    # to do from the response _our_ response parsers give
    # vs. the expected responses in the protocol tests.
    # These are implementation specific changes, not any
    # "we're not following the spec"-type changes.

    # 1. RequestMetadata.  We parse this onto the returned dict, but compliance
    # tests don't have any specs for how to deal with request metadata.
    if 'ResponseMetadata' in parsed:
        del parsed['ResponseMetadata']
    # 2. Binary blob types.  In the protocol test, blob types, when base64
    # decoded, always decode to something that can be expressed via utf-8.
    # This is not always the case.  In python3, the blob type is designed to
    # return a bytes (not str) object.  However, for these tests we'll work for
    # any bytes type, and decode it as utf-8 because we know that's safe for
    # the compliance tests.
    parsed = _convert_bytes_to_str(parsed)
    # 3. We need to expand the event stream object into the list of events
    for key, value in parsed.items():
        if isinstance(value, EventStream):
            parsed[key] = _convert_bytes_to_str(list(value))
            break
    # 4. We parse the entire error body into the "Error" field for rest-xml
    # which causes some modeled fields in the response to be placed under the
    # error key. We don't have enough information in the test suite to assert
    # these properly, and they probably shouldn't be there in the first place.
    if 'Error' in parsed:
        error_keys = list(parsed['Error'].keys())
        for key in error_keys:
            if key not in ['Code', 'Message']:
                del parsed['Error'][key]
    return parsed


def _convert_bytes_to_str(parsed):
    if isinstance(parsed, dict):
        new_dict = {}
        for key, value in parsed.items():
            new_dict[key] = _convert_bytes_to_str(value)
        return new_dict
    elif isinstance(parsed, bytes):
        return parsed.decode('utf-8')
    elif isinstance(parsed, list):
        new_list = []
        for item in parsed:
            new_list.append(_convert_bytes_to_str(item))
        return new_list
    else:
        return parsed


def _compliance_timestamp_parser(value):
    datetime = parse_timestamp(value)
    # Convert from our time zone to UTC
    datetime = datetime.astimezone(tzutc())
    # Convert to epoch.
    return int(timegm(datetime.timetuple()))


def _output_failure_message(
    protocol_type, case, actual_parsed, expected_result, error
):
    j = _try_json_dump
    error_message = (
        "\nDescription           : %s (%s:%s)\n"
        "Protocol:             : %s\n"
        "Given                 : %s\n"
        "Response              : %s\n"
        "Expected serialization: %s\n"
        "Actual serialization  : %s\n"
        "Assertion message     : %s\n"
        % (
            case['description'],
            case['suite_id'],
            case['test_id'],
            protocol_type,
            j(case['given']),
            j(case['response']),
            j(expected_result),
            j(actual_parsed),
            error,
        )
    )
    raise AssertionError(error_message)


def _input_failure_message(protocol_type, case, actual_request, error):
    j = _try_json_dump
    error_message = (
        "\nDescription           : %s (%s:%s)\n"
        "Protocol:             : %s\n"
        "Given                 : %s\n"
        "Params                : %s\n"
        "Expected serialization: %s\n"
        "Actual serialization  : %s\n"
        "Assertion message     : %s\n"
        % (
            case['description'],
            case['suite_id'],
            case['test_id'],
            protocol_type,
            j(case['given']),
            j(case['params']),
            j(case['serialized']),
            j(actual_request),
            error,
        )
    )
    raise AssertionError(error_message)


def _try_json_dump(obj):
    try:
        return json.dumps(obj)
    except (ValueError, TypeError):
        return str(obj)


def assert_equal(first, second, prefix):
    # A better assert equals.  It allows you to just provide
    # prefix instead of the entire message.
    try:
        assert first == second
    except Exception:
        try:
            better = "{} (actual != expected)\n{} !=\n{}".format(
                prefix,
                json.dumps(first, indent=2),
                json.dumps(second, indent=2),
            )
        except (ValueError, TypeError):
            better = "{} (actual != expected)\n{} !=\n{}".format(
                prefix, first, second
            )
        raise AssertionError(better)


def _serialize_request_description(request_dict):
    if isinstance(request_dict.get('body'), dict):
        # urlencode the request body.
        encoded = percent_encode_sequence(request_dict['body']).encode('utf-8')
        request_dict['body'] = encoded
    if isinstance(request_dict.get('query_string'), dict):
        encoded = percent_encode_sequence(request_dict.get('query_string'))
        if encoded:
            # 'requests' automatically handle this, but we in the
            # test runner we need to handle the case where the url_path
            # already has query params.
            if '?' not in request_dict['url_path']:
                request_dict['url_path'] += '?%s' % encoded
            else:
                request_dict['url_path'] += '&%s' % encoded


def _assert_requests_equal(actual, expected):
    assert_equal(
        actual['body'], expected.get('body', '').encode('utf-8'), 'Body value'
    )
    actual_headers = HeadersDict(actual['headers'])
    expected_headers = HeadersDict(expected.get('headers', {}))
    excluded_headers = expected.get('forbidHeaders', [])
    _assert_expected_headers_in_request(
        actual_headers, expected_headers, excluded_headers
    )
    assert_equal(actual['url_path'], expected.get('uri', ''), "URI")
    if 'method' in expected:
        assert_equal(actual['method'], expected['method'], "Method")


def _assert_expected_headers_in_request(actual, expected, excluded_headers):
    for header, value in expected.items():
        assert header in actual
        assert actual[header] == value
    for header in excluded_headers:
        assert header not in actual


def _walk_files():
    # Check for a shortcut when running the tests interactively.
    # If a BOTOCORE_TEST env var is defined, that file is used as the
    # only test to run.  Useful when doing feature development.
    single_file = os.environ.get('BOTOCORE_TEST')
    if single_file is not None:
        yield os.path.abspath(single_file)
    else:
        for root, _, filenames in os.walk(TEST_DIR):
            for filename in filenames:
                yield os.path.join(root, filename)


def _load_cases(full_path):
    # During developement, you can set the BOTOCORE_TEST_ID
    # to run a specific test suite or even a specific test case.
    # The format is BOTOCORE_TEST_ID=suite_id:test_id or
    # BOTOCORE_TEST_ID=suite_id
    suite_id, test_id = _get_suite_test_id()
    all_test_data = json.load(open(full_path), object_pairs_hook=OrderedDict)
    basename = os.path.basename(full_path)
    for i, test_data in enumerate(all_test_data):
        if suite_id is not None and i != suite_id:
            continue
        cases = test_data.pop('cases')
        description = test_data['description']
        for j, case in enumerate(cases):
            if test_id is not None and j != test_id:
                continue
            case['description'] = description
            case['suite_id'] = i
            case['test_id'] = j
            yield (test_data, case, basename)


def _get_suite_test_id():
    if 'BOTOCORE_TEST_ID' not in os.environ:
        return None, None
    test_id = None
    suite_id = None
    split = os.environ['BOTOCORE_TEST_ID'].split(':')
    try:
        if len(split) == 2:
            suite_id, test_id = int(split[0]), int(split[1])
        else:
            suite_id = int(split([0]))
    except TypeError:
        # Same exception, just give a better error message.
        raise TypeError(
            "Invalid format for BOTOCORE_TEST_ID, should be "
            "suite_id[:test_id], and both values should be "
            "integers."
        )
    return suite_id, test_id



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import datetime

import botocore.session
from botocore.compat import parse_qs, urlparse
from botocore.stub import ANY, Stubber
from tests import BaseSessionTest, ClientHTTPStubber, mock, unittest


class TestIdempotencyToken(unittest.TestCase):
    def setUp(self):
        self.function_name = 'purchase_scheduled_instances'
        self.region = 'us-west-2'
        self.session = botocore.session.get_session()
        self.client = self.session.create_client('ec2', self.region)
        self.stubber = Stubber(self.client)
        self.service_response = {}
        self.params_seen = []

        # Record all the parameters that get seen
        self.client.meta.events.register_first(
            'before-call.*.*',
            self.collect_params,
            unique_id='TestIdempotencyToken',
        )

    def collect_params(self, model, params, *args, **kwargs):
        self.params_seen.extend(params['body'].keys())

    def test_provided_idempotency_token(self):
        expected_params = {
            'PurchaseRequests': [
                {'PurchaseToken': 'foo', 'InstanceCount': 123}
            ],
            'ClientToken': ANY,
        }
        self.stubber.add_response(
            self.function_name, self.service_response, expected_params
        )

        with self.stubber:
            self.client.purchase_scheduled_instances(
                PurchaseRequests=[
                    {'PurchaseToken': 'foo', 'InstanceCount': 123}
                ],
                ClientToken='foobar',
            )
            self.assertIn('ClientToken', self.params_seen)

    def test_insert_idempotency_token(self):
        expected_params = {
            'PurchaseRequests': [
                {'PurchaseToken': 'foo', 'InstanceCount': 123}
            ],
        }

        self.stubber.add_response(
            self.function_name, self.service_response, expected_params
        )

        with self.stubber:
            self.client.purchase_scheduled_instances(
                PurchaseRequests=[
                    {'PurchaseToken': 'foo', 'InstanceCount': 123}
                ]
            )
            self.assertIn('ClientToken', self.params_seen)


class TestCopySnapshotCustomization(BaseSessionTest):
    def setUp(self):
        super().setUp()
        self.session = botocore.session.get_session()
        self.client = self.session.create_client('ec2', 'us-east-1')
        self.http_stubber = ClientHTTPStubber(self.client)
        self.snapshot_id = 'snap-0123abc'
        self.copy_response = (
            '<?xml version="1.0" encoding="UTF-8"?>\n'
            '<CopySnapshotResponse>\n'
            '<snapshotId>%s</snapshotId>\n'
            '</CopySnapshotResponse>\n'
        )
        self.now = datetime.datetime(2011, 9, 9, 23, 36)
        self.datetime_patch = mock.patch.object(
            botocore.auth.datetime,
            'datetime',
            mock.Mock(wraps=datetime.datetime),
        )
        self.mocked_datetime = self.datetime_patch.start()
        self.mocked_datetime.utcnow.return_value = self.now

    def tearDown(self):
        super().tearDown()
        self.datetime_patch.stop()

    def add_copy_snapshot_response(self, snapshot_id):
        body = (self.copy_response % snapshot_id).encode('utf-8')
        self.http_stubber.add_response(body=body)

    def test_copy_snapshot_injects_presigned_url(self):
        self.add_copy_snapshot_response(self.snapshot_id)
        with self.http_stubber:
            result = self.client.copy_snapshot(
                SourceRegion='us-west-2',
                SourceSnapshotId=self.snapshot_id,
            )
        self.assertEqual(result['SnapshotId'], self.snapshot_id)
        self.assertEqual(len(self.http_stubber.requests), 1)
        snapshot_request = self.http_stubber.requests[0]
        body = parse_qs(snapshot_request.body)
        self.assertIn('PresignedUrl', body)
        presigned_url = urlparse(body['PresignedUrl'][0])
        self.assertEqual(presigned_url.scheme, 'https')
        self.assertEqual(presigned_url.netloc, 'ec2.us-west-2.amazonaws.com')
        query_args = parse_qs(presigned_url.query)
        self.assertEqual(query_args['Action'], ['CopySnapshot'])
        self.assertEqual(query_args['Version'], ['2016-11-15'])
        self.assertEqual(query_args['SourceRegion'], ['us-west-2'])
        self.assertEqual(query_args['DestinationRegion'], ['us-east-1'])
        self.assertEqual(query_args['SourceSnapshotId'], [self.snapshot_id])
        self.assertEqual(query_args['X-Amz-Algorithm'], ['AWS4-HMAC-SHA256'])
        expected_credential = 'access_key/20110909/us-west-2/ec2/aws4_request'
        self.assertEqual(query_args['X-Amz-Credential'], [expected_credential])
        self.assertEqual(query_args['X-Amz-Date'], ['20110909T233600Z'])
        self.assertEqual(query_args['X-Amz-Expires'], ['3600'])
        self.assertEqual(query_args['X-Amz-SignedHeaders'], ['host'])
        expected_signature = (
            'a94a6b52afdf3daa34c2e2a38a62b72c8dac129c9904c61aa1a5d86e38628537'
        )
        self.assertEqual(query_args['X-Amz-Signature'], [expected_signature])
